"""Utilities for interacting with S3."""

from typing import TYPE_CHECKING, Literal

import boto3

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client as S3ClientType
else:
    S3ClientType = object


# create a singleton s3 client
s3_client: S3ClientType = boto3.client("s3")


def check_experiment_exists(
    s3: S3ClientType, bucket: str, bucket_prefix: str, experiment_id: str
) -> bool:
    """Check if an experiment exists in S3."""
    obj_count = s3.list_objects_v2(
        Bucket=bucket, Prefix=f"{bucket_prefix}/{experiment_id}"
    ).get("KeyCount", 0)
    return obj_count > 0


def raise_on_forbidden_experiment(
    s3: S3ClientType,
    bucket: str,
    bucket_prefix: str,
    experiment_id: str,
    existing_artifacts: Literal["forbid", "overwrite"] = "forbid",
) -> None:
    """Raise an error if an experiment exists in S3."""
    if existing_artifacts == "forbid" and check_experiment_exists(
        s3, bucket, bucket_prefix, experiment_id
    ):
        msg = f"Experiment {experiment_id} already exists (bucket: {bucket}, bucket_prefix: {bucket_prefix})"
        raise ValueError(msg)



"""Filesystem utilities."""

import logging
import shutil
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING

import boto3
import requests
from pydantic import AnyUrl

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client as S3ClientType
else:
    S3ClientType = object

logger = logging.getLogger(__name__)
s3: S3ClientType = boto3.client("s3")


def fetch_uri(  # noqa: C901
    uri: AnyUrl | str,
    local_path: Path,
    use_cache: bool = True,
    logger_fn: Callable = logger.info,
    s3: S3ClientType = s3,
) -> Path:
    """Fetch a file from a uri and return the local path.

    Caching is enabled by default and works by
    checking if the file exists locally before downloading it
    to avoid downloading the same file multiple times.

    Args:
        uri (AnyUrl): The uri to fetch
        local_path (Path): The local path to save the fetched file
        use_cache (bool): Whether to use the cache
        logger_fn (Callable): The logger function to use
        s3 (S3Client): The S3 client to use

    Returns:
        local_path (Path): The local path of the fetched file
    """
    if isinstance(uri, str):
        uri = AnyUrl(uri)
    if uri.scheme == "s3":
        bucket = uri.host
        if not uri.path:
            raise ValueError(f"S3URI:NO_PATH:{uri}")
        if not bucket:
            raise ValueError(f"S3URI:NO_BUCKET:{uri}")
        path = uri.path[1:]
        if not local_path.exists() or not use_cache:
            logger_fn(f"Downloading {uri}...")
            local_path.parent.mkdir(parents=True, exist_ok=True)
            s3.download_file(bucket, path, str(local_path))
        else:
            logger_fn(f"File {local_path} already exists, skipping download.")
    elif uri.scheme == "http" or uri.scheme == "https":
        if not local_path.exists() or not use_cache:
            logger_fn(f"Downloading {uri}...")
            local_path.parent.mkdir(parents=True, exist_ok=True)
            with open(local_path, "wb") as f:
                f.write(requests.get(str(uri), timeout=60).content)
        else:
            logger_fn(f"File {local_path} already exists, skipping download.")
    elif uri.scheme == "file":
        if not local_path.exists() or not use_cache:
            logger_fn(f"Copying {uri} to {local_path}...")
            local_path.parent.mkdir(parents=True, exist_ok=True)
            if uri.path:
                shutil.copy(uri.path, local_path.as_posix())
            else:
                msg = f"File URI:NO_PATH:{uri}"
                logger_fn(msg)
                raise ValueError(msg)
        else:
            logger_fn(f"File {local_path} already exists, skipping copy.")
    else:
        raise NotImplementedError(f"URI:SCHEME:{uri.scheme}")
    return local_path



"""Allocate an experiment to a workflow run."""

import tempfile
from collections.abc import Callable, Sequence
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING

import boto3
import pandas as pd
import yaml
from hatchet_sdk import TaskRunRef, TriggerWorkflowOptions
from pydantic import BaseModel, Field, FilePath
from tqdm import tqdm

from scythe.base import S3Url
from scythe.registry import Standalone, TInput, TOutput
from scythe.scatter_gather import (
    RecursionMap,
    ScatterGatherInput,
    ScatterGatherResult,
    scatter_gather,
)
from scythe.storage import ScytheStorageSettings
from scythe.utils.results import save_and_upload_parquets

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = object

s3 = boto3.client("s3")


class ExperimentSpecsMismatchError(Exception):
    """An error raised when the specs for an experiment do not match the expected type."""

    def __init__(self, expected_type: type, actual_type: type):
        """Initialize the error."""
        self.expected_type = expected_type
        self.actual_type = actual_type
        super().__init__(
            f"Expected type {expected_type.__name__}, got {actual_type.__name__}."
        )


class DuplicateInputArtifactsError(Exception):
    """An error raised when a file is duplicated in the input artifacts."""

    def __init__(self, file_name: str, field_name: str):
        """Initialize the error."""
        self.file_name = file_name
        self.field_name = field_name
        super().__init__(
            f"File with name {file_name} for field {field_name} is duplicated in the "
            f"input artifacts (i.e. multiple files with the same name but in different "
            "directories.) "
        )


class InputArtifactLocations(BaseModel):
    """The locations of the input artifacts for an experiment."""

    files: dict[str, set[S3Url]]


class ExperimentRunManifest(BaseModel):
    """The manifest for an experiment run."""

    workflow_run_id: str
    experiment_id: str
    experiment_name: str
    io_spec: S3Url
    input_artifacts: S3Url | None = None


# TODO: consider factoring this into a an ExperimentRun Class
# which ought to include things like artifact location,
# automatically uploading referenced artifacts, etc.
def allocate_experiment(  # noqa: C901
    experiment_id: str,
    experiment: Standalone[TInput, TOutput],
    specs: Sequence[TInput],
    recursion_map: RecursionMap | None = None,
    construct_filekey: Callable[[str], str] | None = None,
    storage_settings: ScytheStorageSettings | None = None,
    s3_client: S3Client | None = None,
) -> TaskRunRef[ScatterGatherInput, ScatterGatherResult]:
    """Allocate an experiment to a workflow run."""
    s3_client = s3_client or s3
    storage_settings = storage_settings or ScytheStorageSettings()

    mismatching_types = [
        type(spec)
        for spec in specs
        if not isinstance(spec, experiment.config.input_validator)
    ]
    if mismatching_types:
        raise ExperimentSpecsMismatchError(
            expected_type=experiment.config.input_validator,
            actual_type=mismatching_types[0],
        )
    for spec in specs:
        spec.experiment_id = experiment_id

    # handle local file transfer to s3
    # TODO: this will cause a race condition if multile files have the same name but
    # live in different directories.
    # hence the check above
    def construct_input_artifact_key(pth: FilePath, field_name: str):
        return f"{storage_settings.BUCKET_PREFIX}/{experiment_id}/artifacts/{field_name}/{pth.name}"

    local_input_artifact_paths = [
        spec._local_input_artifact_file_paths for spec in specs
    ]
    input_artifacts: dict[str, set[FilePath]] = {}
    at_least_one_input_artifact = False
    for spec_paths in local_input_artifact_paths:
        for field_name, fpath in spec_paths.items():
            input_artifacts.setdefault(field_name, set()).add(fpath)
            at_least_one_input_artifact = True

    all_input_artifact_names: dict[str, set[str]] = {}
    for field_name, fpaths in input_artifacts.items():
        for fpath in fpaths:
            file_key = construct_input_artifact_key(fpath, field_name)
            if file_key in all_input_artifact_names.get(field_name, set()):
                raise DuplicateInputArtifactsError(file_key, field_name)
            all_input_artifact_names.setdefault(field_name, set()).add(file_key)

    input_artifacts_s3_urls = None
    if at_least_one_input_artifact:
        input_artifacts_s3_urls, input_artifacts_s3_url_maps = upload_input_artifacts(
            input_artifacts,
            s3_client,
            storage_settings.BUCKET,
            construct_input_artifact_key,
        )
        for spec in specs:
            for field_name, fpath in spec._local_input_artifact_file_paths.items():
                uri_map = input_artifacts_s3_url_maps[field_name]
                uri = uri_map[fpath]
                setattr(spec, field_name, uri)

    df = pd.DataFrame([s.model_dump(mode="json") for s in specs])
    df_name = "specs"

    def construct_filekey_(filename: str):
        return f"{storage_settings.BUCKET_PREFIX}/{experiment_id}/artifacts/specs/{filename}.pq"

    construct_filekey = construct_filekey or construct_filekey_

    uris = save_and_upload_parquets(
        collected_dfs={df_name: df},
        s3=s3_client,
        bucket=storage_settings.BUCKET,
        output_key_constructor=construct_filekey,
    )
    uri = uris[df_name]
    scatter_gather_input = ScatterGatherInput(
        experiment_id=experiment_id,
        task_name=experiment.name,
        specs_path=uri,
        recursion_map=recursion_map or RecursionMap(path=None, factor=10, max_depth=0),
        storage_settings=storage_settings,
    )

    run_ref = scatter_gather.run_no_wait(
        scatter_gather_input,
        options=TriggerWorkflowOptions(
            additional_metadata={
                "experiment_id": experiment_id,
                "experiment_name": experiment.name,
                "level": 0,
            }
        ),
    )
    workflow_run_id = run_ref.workflow_run_id

    input_validator = experiment.input_validator
    output_validator = experiment._output_validator
    if output_validator is None:
        msg = "Output validator is not set for experiment"
        raise ValueError(msg)

    class ExperimentIO(BaseModel):
        """The input and output schema for the experiment."""

        input: input_validator = Field(  # pyright: ignore [reportInvalidTypeForm]
            default=..., description="The input for the experiment."
        )
        output: output_validator = Field(  # pyright: ignore [reportInvalidTypeForm]
            default=..., description="The output for the experiment."
        )

    schema = ExperimentIO.model_json_schema()

    with tempfile.TemporaryDirectory() as temp_dir:
        tdir = Path(temp_dir)
        manifest_path = tdir / "manifest.yml"
        io_path = tdir / "experiment_io_spec.yml"
        input_artifacts_path = tdir / "input_artifacts.yml"

        # dump and upload the schema
        with open(io_path, "w") as f:
            yaml.dump(schema, f, indent=2)
        io_file_key = (
            f"{storage_settings.BUCKET_PREFIX}/{experiment_id}/experiment_io_spec.yml"
        )
        s3_client.upload_file(
            Bucket=storage_settings.BUCKET,
            Key=io_file_key,
            Filename=io_path.as_posix(),
        )

        # dump and upload the source files
        s3_input_artifacts_key = (
            f"{storage_settings.BUCKET_PREFIX}/{experiment_id}/input_artifacts.yml"
        )
        if input_artifacts_s3_urls:
            with open(input_artifacts_path, "w") as f:
                yaml.dump(
                    input_artifacts_s3_urls.model_dump(mode="json"),
                    f,
                    indent=2,
                )
            s3_client.upload_file(
                Bucket=storage_settings.BUCKET,
                Key=s3_input_artifacts_key,
                Filename=input_artifacts_path.as_posix(),
            )

        manifest_file_key = (
            f"{storage_settings.BUCKET_PREFIX}/{experiment_id}/manifest.yml"
        )
        io_spec_uri = S3Url(f"s3://{storage_settings.BUCKET}/{io_file_key}")
        input_artifacts_uri = S3Url(
            f"s3://{storage_settings.BUCKET}/{s3_input_artifacts_key}"
        )
        manifest = ExperimentRunManifest(
            workflow_run_id=workflow_run_id,
            experiment_id=experiment_id,
            experiment_name=experiment.name,
            io_spec=io_spec_uri,
            input_artifacts=input_artifacts_uri,
        )
        with open(manifest_path, "w") as f:
            yaml.dump(manifest.model_dump(mode="json"), f, indent=2)
        s3_client.upload_file(
            Bucket=storage_settings.BUCKET,
            Key=manifest_file_key,
            Filename=manifest_path.as_posix(),
        )

    return run_ref


def upload_input_artifacts(
    input_artifacts: dict[str, set[FilePath]],
    s3_client: S3Client,
    bucket: str,
    construct_filekey: Callable[[FilePath, str], str],
) -> tuple[InputArtifactLocations, dict[str, dict[FilePath, S3Url]]]:
    """Upload source files to S3."""

    def handle_path(pth: FilePath, field_name: str):
        filekey = construct_filekey(pth, field_name)
        uri = S3Url(f"s3://{bucket}/{filekey}")
        s3_client.upload_file(
            Bucket=bucket,
            Key=filekey,
            Filename=pth.as_posix(),
        )
        return field_name, uri

    args: list[tuple[FilePath, str]] = []
    for field_name, paths in input_artifacts.items():
        for pth in paths:
            args.append((pth, field_name))
    with ThreadPoolExecutor(max_workers=10) as executor:
        first_args = [a[0] for a in args]
        second_args = [a[1] for a in args]
        results = list(
            tqdm(
                executor.map(handle_path, first_args, second_args),
                total=len(args),
                desc="Uploading source files",
            )
        )
    uris: dict[str, set[S3Url]] = {}
    for field_name, uri in results:
        uris.setdefault(field_name, set()).add(uri)
    uri_maps: dict[str, dict[FilePath, S3Url]] = {}
    for (field_name, uri), (pth, _) in zip(results, args, strict=True):
        uri_maps.setdefault(field_name, {})[pth] = uri
    return InputArtifactLocations(files=uris), uri_maps



"""Models for Simulation Specifications."""

import importlib
import importlib.resources
import json
import logging
from pathlib import Path
from typing import Any, cast

import pandas as pd
from pydantic import (
    AnyUrl,
    BaseModel,
    Field,
    FilePath,
    HttpUrl,
    UrlConstraints,
    field_serializer,
    field_validator,
)
from pydantic.json_schema import SkipJsonSchema

from scythe.utils.filesys import fetch_uri
from scythe.utils.results import serialize_df_dict

logger = logging.getLogger(__name__)


# TODO: should experiment ids be uuid trims?
# or should they be human readable (creating unicity issues...)?
# or should they also relate to hatchet auto-generated data?
class BaseSpec(BaseModel, extra="allow", arbitrary_types_allowed=True):
    """A base spec for running a simulation.

    The main features are utilities to fetch files from uris
    and generate a locally scoped path for the files.
    according to the experiment_id.
    """

    experiment_id: str = Field(..., description="The experiment_id of the spec")

    def local_path(self, pth: AnyUrl) -> Path:
        """Return the local path of a uri scoped to the experiment_id.

        Note that this should only be used for non-ephemeral files.

        Args:
            pth (AnyUrl): The uri to convert to a local path

        Returns:
            local_path (Path): The local path of the uri
        """
        path = pth.path

        # get the location of the scythe package
        scythe_dir = importlib.resources.files("scythe")
        # convert the traversable to an absolute path
        scythe_dir = Path(cast(Path, scythe_dir))
        # get the local artifacts directory
        local_artifacts_dir = scythe_dir.parent / "cache"

        if not path or path == "/":
            raise ValueError(f"URI:NO_PATH:{pth}")
        if path.startswith("/"):
            path = path[1:]
        return local_artifacts_dir / self.experiment_id / path

    def log(self, msg: str):
        """Log a message to the context or to the logger.

        Args:
            msg (str): The message to log
        """
        logger.info(msg)

    def fetch_uri(self, uri: AnyUrl, use_cache: bool = True) -> Path:
        """Fetch a file from a uri and return the local path.

        Args:
            uri (AnyUrl): The uri to fetch
            use_cache (bool): Whether to use the cache

        Returns:
            local_path (Path): The local path of the fetched file
        """
        local_path = self.local_path(uri)
        return fetch_uri(uri, local_path, use_cache, self.log)


class ExperimentIndexNotSerializableError(Exception):
    """An error for when an experiment index is not serializable."""

    def __init__(self, cls: type[BaseModel]):
        """Initialize the error."""
        self.cls = cls
        super().__init__(f"Index data is not serializable as json: {cls.__name__}")


class ExperimentIndexAdditionalDataDoesNotMatchNRowsError(Exception):
    """An error for when the additional index data does not match the number of rows."""

    def __init__(
        self, n_rows: int, additional_index_data: dict[str, Any] | list[dict[str, Any]]
    ):
        """Initialize the error."""
        self.n_rows = n_rows
        self.additional_index_data = additional_index_data
        super().__init__(
            f"Additional index data does not match the number of rows: {n_rows} != {len(additional_index_data)}"
        )


class ExperimentIndexAdditionalDataOverlapsWithSpecError(Exception):
    """An error for when the additional index data overlaps with the spec."""

    def __init__(self, overlapping_keys: list[str]):
        """Initialize the error."""
        self.overlapping_keys = overlapping_keys
        super().__init__(
            f"Additional index data overlaps with the spec: {overlapping_keys}"
        )


class S3Url(AnyUrl):
    """A URL for an S3 object."""

    _constraints = UrlConstraints(allowed_schemes=["s3"], host_required=True)


FileReference = S3Url | HttpUrl | Path | FilePath


# TODO: consider making the payload a generic?
class ExperimentInputSpec(BaseSpec):
    """A spec for running a leaf workflow."""

    sort_index: int = Field(..., description="The sort index of the leaf.", ge=0)
    workflow_run_id: str | None = Field(
        default=None, description="The workflow run id of the leaf."
    )
    root_workflow_run_id: str | None = Field(
        default=None, description="The root workflow run id of the leaf."
    )

    def make_multiindex(
        self,
        additional_index_data: dict[str, Any] | list[dict[str, Any]] | None = None,
        n_rows: int = 1,
        include_sort_subindex: bool = True,
    ) -> pd.MultiIndex:
        """Make a MultiIndex from the Spec, and any other methods which might create index data.

        Note that index data should generally be considered as features or inputs, rather than outputs.

        TODO: Feel free to add more args to this method if more values need to be computed.

        Returns:
            multi_index (pd.MultiIndex): The MultiIndex.
        """
        if isinstance(additional_index_data, list) and n_rows != len(
            additional_index_data
        ):
            raise ExperimentIndexAdditionalDataDoesNotMatchNRowsError(
                n_rows, additional_index_data
            )

        index_data: list[dict[str, Any]] = [
            self.model_dump(mode="json") for _ in range(n_rows)
        ]

        if isinstance(additional_index_data, dict):
            if any(k in self.model_dump(mode="json") for k in additional_index_data):
                overlapping_keys = [
                    k
                    for k in additional_index_data
                    if k in self.model_dump(mode="json")
                ]
                raise ExperimentIndexAdditionalDataOverlapsWithSpecError(
                    overlapping_keys
                )
            for d in index_data:
                d.update(additional_index_data)

        elif isinstance(additional_index_data, list):
            for d, ad in zip(index_data, additional_index_data, strict=True):
                if any(k in d for k in ad):
                    overlapping_keys = [k for k in ad if k in d]
                    raise ExperimentIndexAdditionalDataOverlapsWithSpecError(
                        overlapping_keys
                    )
                d.update(ad)

        try:
            json.dumps(index_data)
        except Exception as e:
            raise ExperimentIndexNotSerializableError(self.__class__) from e

        df = pd.DataFrame(index_data)
        if include_sort_subindex and n_rows > 1:
            df["sort_subindex"] = list(range(n_rows))

        return pd.MultiIndex.from_frame(df)

    @classmethod
    def _file_reference_fields(cls) -> list[str]:
        """Get the file reference fields."""
        annotations = cls.model_fields
        return [k for k, v in annotations.items() if v.annotation is FileReference]

    @property
    def _local_input_artifact_file_paths(self) -> dict[str, FilePath]:
        """Get the local source file paths."""
        data = self.model_dump()
        return {
            k: data[k]
            for k in self._file_reference_fields()
            if isinstance(data[k], Path)
        }


class ScalarInDataframesError(Exception):
    """An error for when a scalar is in the dataframes."""

    def __init__(self, scalar: Any):
        """Initialize the error."""
        self.scalar = scalar
        super().__init__("`scalars` key is already in `dataframes`")


class ExperimentOutputSpec(BaseModel, arbitrary_types_allowed=True):
    """A spec for the output of a leaf workflow."""

    # TODO: make this extensible with scalar columns
    dataframes: SkipJsonSchema[dict[str, pd.DataFrame]] = Field(default_factory=dict)

    @field_validator("dataframes", mode="before")
    def validate_dataframes(cls, v):
        """Validate the dataframes via deserialization."""
        if isinstance(v, dict):
            v = {
                k: (
                    v
                    if isinstance(v, pd.DataFrame)
                    else pd.DataFrame.from_dict(v, orient="tight")
                )
                for k, v in v.items()
            }
        return v

    @field_serializer("dataframes")
    def serialize_dataframes(self, v):
        """Serialize the dataframes via serialization."""
        return serialize_df_dict(v)

    def add_scalars(self, input_spec: ExperimentInputSpec):
        """Update the dataframes with the input spec."""
        scalar_data = self.model_dump(mode="json", exclude={"dataframes"})
        if len(scalar_data) == 0:
            return
        results = [scalar_data]
        multi_index = input_spec.make_multiindex(
            results, n_rows=1, include_sort_subindex=False
        )
        df = pd.DataFrame(results, index=multi_index)
        if "scalars" in self.dataframes:
            raise ScalarInDataframesError(self.dataframes["scalars"])
        self.dataframes["scalars"] = df
